# Overview:
# cal2jd
# calculateEOF
# calculatePC
# catchmentAverage
# fitIndex
# fitTrend
# fitTrend2D
# jd2cal
# jd2yr
# selectPolygon
# showData
# showEigenvalues
# showEOF
# showEOFlocal
# showPCs
# showTimeseries
# showTimeseries_DSI
# time2decimalYears

import numpy as np
from matplotlib.path import Path
from scipy.signal import hilbert
import matplotlib.pyplot as plt
from matplotlib.path import Path
from scipy.io import loadmat
from datetime import datetime
import matplotlib.dates as mdates


def cal2jd(yr, mn, dy):
    """CAL2JD  Converts calendar date to Julian date using algorithm
      from "Practical Ephemeris Calculations" by Oliver Montenbruck
      (Springer-Verlag, 1989). Uses astronomical year for B.C. dates
      (2 BC = -1 yr). Non-vectorized version. See also DOY2JD, GPS2JD,
      JD2CAL, JD2DOW, JD2DOY, JD2GPS, JD2YR, YR2JD.
    Version: 1999-04-24
    Usage:   jd=cal2jd(yr,mn,dy)
    Input:   yr - calendar year (4-digit including century)
             mn - calendar month
             dy - calendar day (including factional day)
    Output:  jd - jJulian date

    Copyright (c) 2011, Michael R. Craymer
    All rights reserved.
    Email: mike@craymer.com"""
    if len(locals()) != 3:
        raise ValueError("Incorrect number of input arguments")
    if not (1 <= mn <= 12):
        raise ValueError("Invalid input month")

    if dy < 1:
        if (mn == 2 and dy > 29) or (mn in [4, 6, 9, 11] and dy > 30) or (dy > 31):
            raise ValueError("Invalid input day")

    if mn > 2:
        y = yr
        m = mn
    else:
        y = yr - 1
        m = mn + 12

    date1 = 4 + 31 * (10 + 12 * 1582)  # Last day of Julian calendar (1582.10.04)
    date2 = 15 + 31 * (10 + 12 * 1582)  # First day of Gregorian calendar (1582.10.15)
    date = dy + 31 * (mn + 12 * yr)

    if date <= date1:
        b = -2
    elif date >= date2:
        b = y // 400 - y // 100
    else:
        raise ValueError("Dates between October 5 & 15, 1582 do not exist")

    if y > 0:
        jd = int(365.25 * y) + int(30.6001 * (m + 1)) + b + 1720996.5 + dy
    else:
        jd = int(365.25 * y - 0.75) + int(30.6001 * (m + 1)) + b + 1720996.5 + dy

    return jd


def calculateEOF(Y):
    """
    Calculation of EOFs from a given data matrix Y. The data matrix Y has
    to be centered.
    --Input--
      Y               matrix containing the time series of gridded data sets
                      (n x p)
    --Output--
      eigenvalues     vector containing the p non-zero eigenvalues of the
                      covariance matrix C, sorted according to decreasing
                      size (p x 1)
      E               matrix containing in its columns the eigenvectors
                      (EOFs) of the covariance matrix C. EOFs are sorted
                      according to size of corresponding eigenvalue (n x p).

    Author: Anne Springer
    Date:  2020-09-16"""
    # Calculate Covariance Matrix (with considering the factor 1/n)
    C = Y.T @ Y

    # Calculate matrix of eigenvectors E and eigenvalues lambda
    lambda_, E = np.linalg.eig(C)

    # Sort eigenvalues and eigenvectors according to the size of the eigenvalues
    index = np.argsort(-lambda_)
    lambda_ = lambda_[index]
    E = E[:, index]
    # Calculate eigenvectors of Y*Y'
    E = Y @ E
    # Normalize eigenvectors => EOFs
    E = E @ np.diag(1 / np.sqrt(np.sum(E**2, axis=0)))

    # Optional:
    E = E * np.sign(np.sum(E, axis=0))

    # Eigenvalues
    eigenvalues = lambda_

    return eigenvalues, E


def calculatePC(E, Y):
    """Calculation of principle components (PCs) by projecting the original data
    onto the basis of the EOFs.
    --Input--
        Y               matrix containing the time series of gridded data sets
                        (n x p)
        E               matrix containing in its columns the eigenvectors
                        (EOFs) of the covariance matrix C. EOFs are sorted
                        according to size of corresponding eigenvalue (n x p).

    --Output--
        D               matrix containing the principle components in its row
                        (p x p)

    Author: Anne Springer
    Date:  2020-09-16"""
    D = E.T @ Y
    return D


def catchmentAverage(data, lon, lat, polygon):
    """
    Computation of a catchment averaged time series for a given polygon

    --Input--
        data            Matrix containing the data values (n x m x p)
        lon             longitude values of the grid points (n x 1)
        lat             latitude values of the grid points (m x 1)
        polygon         lon and lat of polygon (bx2)

    --Output--
        twsa              catchment averaged time series (p x 1)

    Author: Anne Springer
    Date:  2020-09-16"""

    # Regular grid
    Lon, Lat = np.meshgrid(lon, lat[::-1])

    # Get values inside polygon
    points = np.vstack((Lon.flatten(), Lat.flatten())).T
    path = Path(polygon)
    inside = path.contains_points(points).reshape(Lon.shape)

    # Compute area
    dLambda = (lon[1] - lon[0]) * np.pi / 180
    dPhi = (lat[1] - lat[0]) * np.pi / 180
    phi = Lat[inside] * np.pi / 180
    areaVec = dLambda * 2.0 * np.sin(dPhi / 2.0) * np.cos(phi)

    # Compute catchment average
    twsa = np.full(data.shape[2], np.nan)
    for i in range(data.shape[2]):
        temp = data[:, :, i]
        twsa[i] = np.sum(temp[inside] * areaVec) / np.sum(areaVec)

    return twsa


def fitIndex(time, data, index, mod):
    """
    Fit trend, annual and semiannual signal, and a selected climate index

    --Input--
        time          time in decimal years (p x 1)
        data          data (n x m x p), n... #lon, m... #lat, p...#timesteps
        index         given climate index (n x 1)
        mod           choose 'annual' if only the annual signal should be
                        included in the parameter estimation, choose 'semiannual'
                        if additionally semi-annual signals should be included

    --Output--
        param         estimated parameters (n x m x #param)
                        Parameters: bias, trend, two components of the annual
                        signal, two components of the semiannual signal,
                        real-valued and imaginary component of the climate index
        dataApprox    data approximated by the estimated parameters (n x m x p)

    Author: Anne Springer
    Date:  2020-09-16
    """
    # Reshape data
    j, k, l = data.shape
    omega = 2 * np.pi
    data = data.reshape(j * k, l)

    # Functional model
    if mod == "annual":  # only annual
        A = np.column_stack(
            [
                np.ones(l),
                time,
                np.cos(omega * time),
                np.sin(omega * time),
                index,
                np.imag(hilbert(index, axis=0)),
            ]
        )
    elif mod == "semiannual":  # annual and semiannual
        A = np.column_stack(
            [
                np.ones(l),
                time,
                np.cos(omega * time),
                np.sin(omega * time),
                np.cos(2 * omega * time),
                np.sin(2 * omega * time),
                index,
                np.imag(hilbert(index, axis=0)),
            ]
        )
    else:
        raise ValueError("Invalid value for 'mod'. Choose 'annual' or 'semiannual'.")

    # Estimate parameters
    param = np.linalg.lstsq(A, data.T, rcond=None)[0]
    dataApprox = A @ param
    param = param.T
    dataApprox = dataApprox.T

    # Reshape data
    param = param.reshape(j, k, param.shape[1])
    dataApprox = dataApprox.reshape(j, k, l)

    return param, dataApprox


def fitTrend(time, data, mod):
    """
    Compute trend, annual and semiannual signal

    --Input--
        time          time in decimal years (p x 1)
        data          data (p x 1), p...#timesteps
        mod           choose 'annual' if only the annual signal should be
                        included in the parameter estimation, choose 'semiannual'
                        if additionally semi-annual signals should be included

    --Output--
        param         estimated parameters (#param x 1)
                        Parameters: bias, trend, two components of the annual
                        signal, two components of the semiannual signal.
        dataApprox    data approximated by the estimated parameters (p x 1)

    Author: Anne Springer
    Date:  2020-09-16
    """
    l = len(data)
    omega = 2 * np.pi

    # Functional model
    if mod == "annual":  # only annual
        A = np.column_stack(
            [np.ones(l), time, np.cos(omega * time), np.sin(omega * time)]
        )
    elif mod == "semiannual":  # annual and semiannual
        A = np.column_stack(
            [
                np.ones(l),
                time,
                np.cos(omega * time),
                np.sin(omega * time),
                np.cos(2 * omega * time),
                np.sin(2 * omega * time),
            ]
        )
    else:
        raise ValueError("Invalid value for 'mod'. Choose 'annual' or 'semiannual'.")

    # Estimate parameters
    param = np.linalg.lstsq(A, data, rcond=None)[0]
    dataApprox = A @ param

    return param, dataApprox


def fitTrend2D(time, data, mod):
    """Compute trend, annual and semiannual signal

    --Input--
        time          time in decimal years (p x 1)
        data          data (n x m x p), n... #lon, m... #lat, p...#timesteps
        mod           choose 'annual' if only the annual signal should be
                        included in the parameter estimation, choose 'semiannual'
                        if additionally semi-annual signals should be included

    --Output--
        param         estimated parameters (n x m x #param)
                        Parameters: bias, trend, two components of the annual
                        signal, two components of the semiannual signal.
        dataApprox    data approximated by the estimated parameters (n x m x p)

    Author: Anne Springer
    Date:  2020-09-16"""
    # Reshape data
    j, k, l = data.shape
    omega = 2 * np.pi
    data = data.reshape(j * k, l)

    # Functional model
    if mod == "annual":  # only annual
        A = np.column_stack(
            [np.ones(l), time, np.cos(omega * time), np.sin(omega * time)]
        )
    elif mod == "semiannual":  # annual and semiannual
        A = np.column_stack(
            [
                np.ones(l),
                time,
                np.cos(omega * time),
                np.sin(omega * time),
                np.cos(2 * omega * time),
                np.sin(2 * omega * time),
            ]
        )
    else:
        raise ValueError("Invalid value for 'mod'. Choose 'annual' or 'semiannual'.")

    # Estimate parameters
    param = np.linalg.lstsq(A, data.T, rcond=None)[0]
    dataApprox = A @ param
    param = param.T
    dataApprox = dataApprox.T

    # Reshape data
    param = param.reshape(j, k, param.shape[1])
    dataApprox = dataApprox.reshape(j, k, l)

    return param, dataApprox


def jd2cal(jd):
    """
    JD2CAL  Converts Julian date to calendar date using algorithm
        from "Practical Ephemeris Calculations" by Oliver Montenbruck
        (Springer-Verlag, 1989). Must use astronomical year for B.C.
        dates (2 BC = -1 yr). Non-vectorized version. See also CAL2JD,
        DOY2JD, GPS2JD, JD2DOW, JD2DOY, JD2GPS, JD2YR, YR2JD.
    Version: 24 Apr 99
    Usage:   [yr, mn, dy]=jd2cal(jd)
    Input:   jd - Julian date
    Output:  yr - year of calendar date
            mn - month of calendar date
            dy - day of calendar date (including decimal)

    Copyright (c) 2011, Michael R. Craymer
    All rights reserved.
    Email: mike@craymer.com
    """
    if len(locals()) != 1:
        raise ValueError("Incorrect number of input arguments")
    if jd < 0:
        raise ValueError("Julian date must be greater than or equal to zero")

    a = int(jd + 0.5)
    if a < 2299161:
        c = a + 1524
    else:
        b = int((a - 1867216.25) / 36524.25)
        c = a + b - int(b / 4) + 1525

    d = int((c - 122.1) / 365.25)
    e = int(365.25 * d)
    f = int((c - e) / 30.6001)
    dy = c - e - int(30.6001 * f) + (jd + 0.5) % 1
    mn = f - 1 - 12 * int(f / 14)
    yr = d - 4715 - int((7 + mn) / 10)

    return yr, mn, dy


def jd2yr(jd):
    """
    JD2YR  Converts Julian date to year and decimal of year.
    . Non-vectorized version. See also CAL2JD, DOY2JD,
        GPS2JD, JD2CAL, JD2DOW, JD2DOY, JD2GPS, JD2YR.
    Version: 24 Apr 99
    Usage:   yr=jd2yr(jd)
    Input:   jd - Julian date
    Output:  yr - year and decimal of year

    Copyright (c) 2011, Michael R. Craymer
    All rights reserved.
    Email: mike@craymer.com
    """
    if len(locals()) != 1:
        raise ValueError("Incorrect number of input arguments")
    if jd < 0:
        raise ValueError("Julian date must be greater than or equal to zero")

    [iyr, mn, yr] = jd2cal(jd)
    jd0 = cal2jd(iyr, 1, 1)
    jd1 = cal2jd(iyr + 1, 1, 1)
    yr = iyr + (jd - jd0) / (jd1 - jd0)

    return yr


def selectPolygon(lon, lat, polygon):
    """
    Select longitudes and latitudes located within or on the border of a polygon

    --Input--
    lon             longitudes (n x 1)
    lat             latitudes  (m x 1)
    polygon         lon and lat of polygon (b x 2)

    --Output--
    IN              index of data points within or on the border of the polygon (n*m x 1)
    Lon             Longitudes corresponding to the data points within the
                    polygon (c x 1)
    Lat             Latitudes corresponding to the data points within the
                    polygon (c x 1)

    Author: Anne Springer
    Date:  2020-09-16
    """

    Lon, Lat = np.meshgrid(lon, lat)

    Lon = Lon.T.flatten()
    Lat = Lat.T.flatten()

    polygon_path = Path(polygon)

    In = polygon_path.contains_points(np.vstack((Lon, Lat)).T, radius=1e-9)

    return In, Lon[In], Lat[In]


def showData(
    data,
    longitude,
    latitude,
    titleString,
    label="mm",
    clim=np.nan,
    cmap="jet",
    reverse=False,
):
    """
    Visualization of (global) gridded data sets.

    --Input--
    data            matrix containing the data values (n x m)
    longitude       longitude values of the grid points (n x 1)
    latitude        latitude values of the grid points (m x 1)
    titleString     string with title
    label           string with label
    clim            2x1 vector with lower and upper limit of colorbar, e.g.
                    [-100 100]
    cmap            choice of colormap: e.g. jet, parula
    reverse         boolean: if reverse==1 invert colormap

    --Output--
    2D-plot of the gridded values

    Author: Anne Springer
    Date:  2020-09-16"""

    if data.ndim == 1:
        data = data.reshape(len(latitude), len(longitude), order="F")

    # Flip data array for plot
    data = data[::-1, :]

    # Convert longitude and latitude to 1D arrays if necessary
    longitude = longitude.squeeze()
    latitude = latitude.squeeze()

    # Plot properties
    fs = 12
    plt.figure(figsize=(10, 6))
    plt.gca().set_position([0.15, 0.15, 0.7, 0.7])
    plt.gca().tick_params(labelsize=fs)

    # Load coast data
    coast = loadmat("data_Lab3_Lab4/coast30.mat")
    lam = coast["lam"].flatten()
    phi = coast["phi"].flatten()
    lam[lam >= 180] -= 360

    # Plot
    plt.imshow(
        data,
        extent=[min(longitude), max(longitude), min(latitude), max(latitude)],
        cmap=cmap if not reverse else plt.cm.get_cmap(cmap).reversed(),
        origin="lower",
    )
    plt.colorbar(label=label)
    plt.plot(lam, phi, "k", linewidth=1)
    plt.axis([-180, 180, -90, 90])
    if clim is not None:
        plt.clim(clim)
    plt.title(titleString)
    plt.xlabel("Longitude")
    plt.ylabel("Latitude")
    # plt.show()


def showEigenvalues(eigenvalues, titleString):
    """
    Visualization of the evolution of the eigenvalues.

    --Input--
    eigenvalues              vector of eigenvalues (p x 1)
    titleString              string with title

    --Output--
    Plot of the eigenvalues.

    Author: Anne Springer
    Date:  2020-09-16
    """
    plt.figure(figsize=(10, 6))
    plt.gca().tick_params(labelsize=12)

    plt.plot(eigenvalues, "-bs", linewidth=2, markersize=5)  # marker="o")
    plt.plot(eigenvalues, "-r", linewidth=2)

    plt.title(f"{titleString} Eigenvalues")
    plt.xlabel("Index")
    plt.ylabel("Eigenvalue")
    # plt.show()


def showEOF(Eof, longitude, latitude, i, titleString):
    """
    Visualization of the spatial pattern of one specific EOF with the index i.

    --Input--
    Eof             vector containing the gridded values of the one EOF to
                    be plotted (n*m x 1, e.g. 64800 x 1)
    longitude       longitude values of the grid points (n x 1, e.g. 360 x 1)
    latitude        latitude values of the grid points (m x 1, e.g. 180 x 1)
    i               index of the EOF to be visualized (EOFs are sorted
                    according to size of eigenvalue)
    titleString     string with title

    --Output--
    Plot of the spatial pattern.

    Author: Anne Springer
    Date:  2020-09-16
    """
    eof = Eof.reshape(180, 360, order="F")

    fs = 12  # Fontsize
    plt.figure(figsize=(8, 6))
    plt.gca().set_position([0.15, 0.15, 0.7, 0.7])
    plt.gca().tick_params(labelsize=fs)

    # Load coast data
    coast = loadmat("data_Lab3_Lab4/coast30.mat")
    lam = coast["lam"].flatten()
    phi = coast["phi"].flatten()
    lam[lam >= 180] -= 360

    # Plot
    plt.imshow(
        eof,
        extent=[min(longitude), max(longitude), min(latitude), max(latitude)],
        cmap="jet_r",
    )
    plt.colorbar()
    plt.plot(lam, phi, "k", linewidth=2)
    plt.axis([-180, 180, -90, 90])
    plt.title(f"{titleString} - EOF {i}")
    plt.xlabel("Longitude")
    plt.ylabel("Latitude")
    plt.clim([-0.04, 0.04])
    # plt.show()


def showEOFlocal(Eof, longitude, latitude, border, i, titleString):
    """
    Visualization of the spatial pattern of one specific EOF with the index i
    on a spatial domain limited by the polygon border.
    Attention: longitude and latitude must have the same number of entries
    corresponding to the number of grid points within the polygon

    --Input--
    Eof             vector containing the gridded values of the one EOF to
                    be plotted (n x 1)
    longitude       longitude values of the grid points (n x 1)
    latitude        latitude values of the grid points (n x 1)
    border          matrix which contains the polygon around the regional
                    area
    i               index of the EOF to be visualized (EOFs are sorted
                    according to size of eigenvalue)
    titleString     string with title

    --Output--
    Plot of the spatial pattern for a regional area.

    Author: Anne Springer
    Date:  2020-09-16
    """

    if len(longitude) == 0 or len(latitude) == 0:
        raise ValueError("Longitude and latitude arrays must not be empty")
    x = np.arange(min(longitude), max(longitude) + 1)
    y = np.arange(max(latitude), min(latitude) - 1, -1)
    X, Y = np.meshgrid(x, y)
    dataRect = np.full(X.shape, np.nan).T

    polygon_path = Path(border)
    In = (
        polygon_path.contains_points(
            np.vstack((X.flatten(), Y.flatten())).T, radius=1e-9
        )
        .reshape(X.shape)
        .T
    )

    dataRect[In] = Eof.real
    dataRect = dataRect.T
    margin = 5

    fs = 12  # Fontsize
    plt.figure(figsize=(8, 6))
    plt.gca().set_position([0.15, 0.15, 0.7, 0.7])
    plt.gca().tick_params(labelsize=fs)

    # Load coast data
    coast = loadmat("data_Lab3_Lab4/coast30.mat")
    lam = coast["lam"].flatten()
    phi = coast["phi"].flatten()
    lam[lam >= 180] -= 360

    # Plot
    im = plt.imshow(
        dataRect,
        extent=[
            min(border[:, 0]),
            max(border[:, 0]),
            min(border[:, 1]),
            max(border[:, 1]),
        ],
        cmap="jet_r",
    )
    plt.colorbar(im)
    im.set_clim(-0.2, 0.2)
    plt.plot(lam, phi, "k", linewidth=2)
    plt.plot(border[:, 0], border[:, 1], "-r", linewidth=2)
    plt.axis(
        [
            min(border[:, 0]) - margin,
            max(border[:, 0]) + margin,
            min(border[:, 1]) - margin,
            max(border[:, 1]) + margin,
        ]
    )

    plt.title(f"{titleString} - EOF {i}")
    plt.xlabel("Longitude")
    plt.ylabel("Latitude")
    plt.show()


def showPCs(PCs, time, i, legendStrings):
    """
    Visualization of the principle component with the index i.

    --Input--
    PCs             matrix containing the principle components which shall
                    be visualized (p x u), with p... number of time steps
                    and u... number of components/models
    time            matrix with [years months] (p x 2)
    i               number of the PC to be visualized (PCs are sorted
                    according to the size of the eigenvalues)
    legendStrings   legend entries {'entry1', 'entry2'}

    --Output--
    Plot of the principle component.

    Author: Anne Springer
    Date:  2020-09-16
    """

    # Convert time to datetime objects
    time_ = [datetime(year, month, 15) for year, month in time]

    fs = 12  # Fontsize
    ms = 20  # MarkerSize
    lw = 2  # LineWidth

    plt.figure(figsize=(16, 6))
    plt.gca().tick_params(labelsize=fs)
    plt.gca().xaxis.set_major_formatter(mdates.DateFormatter("%Y"))
    plt.gca().xaxis.set_major_locator(mdates.YearLocator())
    plt.gca().set_position([0.1, 0.1, 0.8, 0.8])
    plt.gcf().autofmt_xdate()
    plt.gca().set_title(f"PC {i}", fontsize=fs)
    plt.xlabel("Year", fontsize=fs)
    plt.ylabel("[mm]", fontsize=fs)

    if PCs.ndim == 1:
        PCs = PCs.reshape(len(PCs), 1)

    for j in range(PCs.shape[1]):
        plt.plot(
            time_,
            PCs[:, j],
            markersize=ms,
            linewidth=lw,
            label=legendStrings[j],
        )

    plt.legend(loc="best")
    # plt.show()


def showTimeseries(data, ylim, legendStrings, titleString, ylabelString):
    """
    Visualization of time series

    --Input--
    data            list of lists containing the time and data values
                    [[years1, months1, data1], [years2, months2, data2], ...]
    ylim            lower and upper limit of the y-axis (2 x 1)
    legendStrings   legend entries ['entry1', 'entry2']
    titleString     string with title
    ylabelString    string with label of the y-axis

    --Output--
    Plot of the time series

    Author: Anne Springer
    Date:  2020-09-16
    """

    fs = 16  # Fontsize
    ms = 15  # MarkerSize
    lw = 2  # LineWidth

    plt.figure(figsize=(12, 6))
    plt.gca().tick_params(labelsize=fs)
    plt.ylim(ylim)
    plt.gca().xaxis.set_major_formatter(mdates.DateFormatter("%m/%y"))
    plt.gca().xaxis.set_major_locator(mdates.YearLocator())
    plt.gcf().autofmt_xdate()
    plt.gca().set_title(titleString, fontsize=fs)
    plt.xlabel("Time", fontsize=fs)
    plt.ylabel(ylabelString, fontsize=fs)
    plt.gca().set_position([0.1, 0.1, 0.8, 0.8])
    plt.gca().tick_params(labelsize=fs)

    min_t = datetime.max
    max_t = datetime.min

    for entry in data:
        years, months, values = (
            entry[:, 0].astype(int),
            entry[:, 1].astype(int),
            entry[:, 2],
        )
        times = [datetime(year, month, 1) for year, month in zip(years, months)]

        if min(times) < min_t:
            min_t = min(times)
        if max(times) > max_t:
            max_t = max(times)

        # Create a full range of dates from start to end
        full_range = [
            datetime(year, month, 1)
            for year in range(min(years), max(years) + 1)
            for month in range(1, 13)
        ]
        full_range = [
            date for date in full_range if date >= min(times) and date <= max(times)
        ]

        # Fill missing data with NaN
        full_values = np.full(len(full_range), np.nan)
        for t, v in zip(times, values):
            idx = full_range.index(t)
            full_values[idx] = v

        plt.plot(full_range, full_values, ".-", markersize=ms, linewidth=lw)

    plt.legend(legendStrings, loc="best")


def showTimeseries_DSI(data, ylim, legendStrings, titleString, ylabelString):
    """
    Visualization of time series

    --Input--
    data            list of lists containing the time and data values
                    [[years1, months1, data1], [years2, months2, data2], ...]
    ylim            lower and upper limit of the y-axis (2 x 1)
    legendStrings  legend entries ['entry1', 'entry2']
    titleString     string with title
    ylabelString    string with label of the y-axis

    --Output--
    Plot of the time series with additional lines

    Author: Alicia Daubenspeck
    Date:  2024-11-06
    """
    fs = 16  # Fontsize
    ms = 15  # MarkerSize
    lw = 2  # LineWidth

    plt.figure(figsize=(12, 6))
    plt.gca().tick_params(labelsize=fs)
    plt.ylim(ylim)
    plt.gca().xaxis.set_major_formatter(mdates.DateFormatter("%m/%y"))
    plt.gca().xaxis.set_major_locator(mdates.YearLocator())
    plt.gcf().autofmt_xdate()
    plt.gca().set_title(titleString, fontsize=fs)
    plt.xlabel("Time", fontsize=fs)
    plt.ylabel(ylabelString, fontsize=fs)
    plt.gca().set_position([0.1, 0.1, 0.8, 0.8])
    plt.gca().tick_params(labelsize=fs)

    min_t = datetime.max
    max_t = datetime.min

    for i, entry in enumerate(data):
        years, months, values = (
            entry[:, 0].astype(int),
            entry[:, 1].astype(int),
            entry[:, 2],
        )
        times = [datetime(year, month, 1) for year, month in zip(years, months)]

        if min(times) < min_t:
            min_t = min(times)
        if max(times) > max_t:
            max_t = max(times)

        # Create a full range of dates from start to end
        full_range = [
            datetime(year, month, 1)
            for year in range(min(years), max(years) + 1)
            for month in range(1, 13)
        ]
        full_range = [
            date for date in full_range if date >= min(times) and date <= max(times)
        ]

        # Fill missing data with NaN
        full_values = np.full(len(full_range), np.nan)
        for t, v in zip(times, values):
            idx = full_range.index(t)
            full_values[idx] = v

        plt.plot(
            full_range,
            full_values,
            ".-",
            markersize=ms,
            linewidth=lw,
            label=legendStrings[i],
        )

    # Additional lines and legend
    plt.hlines(y=-1.3, xmin=min_t, xmax=max_t, colors="k", label="severe")
    plt.hlines(y=-1.6, xmin=min_t, xmax=max_t, colors="gray", label="extreme")
    plt.hlines(y=-2, xmin=min_t, xmax=max_t, colors="lightgray", label="exceptional")
    plt.savefig("pics_Lab4/ex5/dsi.png", dpi=600)

    plt.legend(loc="best")


def time2decimalYears(years, months):
    """
    Convert time given as years and months to decimal years

    --Input--
        years           vector with years (p x 1)
        months          vector with months(p x 1)

    --Output--
        t               vector with time in decimal years (p x 1)

    Author: Anne Springer
    Date:  2020-09-16"""
    t = np.zeros(len(years))
    for i in range(len(years)):
        jd = cal2jd(years[i], months[i], 1)
        t[i] = jd2yr(jd)

    return t
